#include <cstdlib>
#include <cmath>
#include <ctime>
#include <iostream>

#ifdef _WIN32
#include <windows.h>
#include <sstream>
#endif

#include "Clock.h"
#include "vgalib.h"

using namespace std;

int MAXX;
int MAXY;

extern Uint8 *buffer;
extern SDL_Surface *surface;

int main(int argc, char *argv[])
{
    switch (argc) {
    case 3:
        MAXX = atoi(argv[1]);
        MAXY = atoi(argv[2]);
	MAXX = 16*(MAXX/16);
	MAXY = 16*(MAXY/16);
        break;
    default:
        MAXX = 640;
        MAXY = 480;
        break;
    }

    // Graphics window and palette setup
    init256();
    pal();

    SDL_EnableUNICODE(1);

    Clock watch;
    Uint32 framesDrawn = 0;

    float y[2][1024];

    memset(&y[0][0], 0, sizeof(y[0]));
    memset(&y[1][0], 0, sizeof(y[1]));
    
    int idx = 0;

    // Main loop
    while (1) {

	SDL_PumpEvents();

	// Abort if ESC hit
	Uint8 *keys = SDL_GetKeyState(NULL);
	if (keys[SDLK_ESCAPE])
	    break;
	 
// The theory behind this calculation is simple.
// Assume we have a wave function, f(x), (x=horizontal pixel) as it is at time (n).
// We want to find how the values will change at time (n+1).
//
// Consider the individual pixels to represent the water molecules.
// Assume that each molecule is influenced only by its two neighbours,
// as if they are connected to it with "springs":
//
//      xxx                   xxx
//     x    x                R
//    x      x              O._
//            x            L |\
//  ---------------------------\------
//              x        x      \
//               x      x        `---- This point (O) is only influenced
//                x    x               by the forces of two "springs":
//                 xxxx                the one attached to its left
//                                     neighbour(L) and the one attached
//                                     to its right neighbour(R)
//
// Based on simple Neutonian laws, we will have for the point O:
// (p=position, v=velocity, a=acceleration)
//
//
// p    = p  + v        (new position = old position + velocity)
//  n+1    n    n
//
//
// v  = v    + a        (velocity = old velocity + old acceleration)
//  n    n-1    n-1
//
//                      (springs cause acceleration linear to distance)
//
// a    = k * (pL    - p    ) + k * (pR    - p    )  (k is the spring coeff)
//  n-1          n-2    n-2            n-2    n-2
//
//
// This would work, but would require quite a lot of calculations...
// Instead, we will use the acceleration at time n+1 (instead of n-1),
// which basically means that we will delay the effects of the springs
// for two time slots - as if the springs have a "delayed" effect.
//
// We will also set k=1. Look at what happens (we replace v  first):
//                                                         n
//
//
// p    = p  + v    + a    = (aprox) p  + (p  - p   )  + a
//  n+1    n    n-1    n-1            n     n    n-1      n+1
//
//      = 2*p  - p    + pL    - p   + pR    - p
//           n    n-1     n      n      n      n
//
//      = pL  + pR  - p
//          n     n    n-1
//
// So the end result is amazingly simple: we add the values of the two
// neighbouring pixels, and subtract the old value we used to have!
//
	float *yOldest   = &y[idx][0];
	idx = (idx+1) % 2;
	float *yOld = &y[idx][0];
	int i=0;

	for(i=1;i<MAXX-1; i++) {
	    if (fabs(yOld[i])<MAXY/2)
		qplot(i, MAXY/2 + int(yOld[i]), 0);
	    float val = yOld[i-1] + yOld[i+1] - yOldest[i];
	    yOldest[i] = (short) (0.9999*val);
	    if (fabs(yOldest[i])<MAXY/2)
		qplot(i, MAXY/2 + int(yOldest[i]), 255);
	}
	if (keys[SDLK_SPACE]) {
	    if (fabs(yOldest[MAXX/4])<MAXY/2)
		qplot(MAXX/4, MAXY/2 + int(yOldest[MAXX/4]), 0);
	    yOldest[MAXX/4] = MAXY/2;
	}

        showpage();
	framesDrawn++;
	SDL_Delay(10);
    }
#ifdef _WIN32
    stringstream speed;
    speed << "Rendering " << framesDrawn << " frames in " << (watch.readMS()/1000.0) << " seconds. (" << framesDrawn/(watch.readMS()/1000.0) << " fps)\n";
    MessageBoxA(0, (LPCSTR) speed.str().c_str(), (const char *)"Speed of rendering", MB_OK);
#else
    cout << "Rendering " << framesDrawn << " frames in ";
    cout << (watch.readMS()/1000.0) << " seconds. (";
    cout << framesDrawn/(watch.readMS()/1000.0) << " fps)\n";
#endif
    return 0;
}
